#include <stdint.h>
#include "LCD.h"
//This module requires STM32F10x Std Peripheral Library
#include "stm32f10x.h"

#include "libnsbmp.h"

#if !defined(LCD_ORIENTATION_VERTICAL) && !defined(LCD_ORIENTATION_HORIZONTAL)
#error Define display orientation!
#endif
#if defined(LCD_ORIENTATION_VERTICAL) && defined(LCD_ORIENTATION_HORIZONTAL)
#error Only one display orientation should be defined!
#endif

//==============================================================================
//                  Module's macros
//==============================================================================

//Helper to set gamma register values
#define GREGVAL(byte1, byte2) (uint16_t)(((byte1) << 8) | ((byte2) & 0xFF))
//Helper for bit mask
#define b(n) (1<<(n))

//==============================================================================
//                  Module's internal types
//==============================================================================

//IL9325 bits
enum
{
    //Reg 01h Driver Output control
    SS=b(8), SM=b(10),
    //Reg 02h LCD Driving control
    EOR=b(8), BC0=b(9),
    //Reg 03h Entry mode
    AM=b(3), ID0=b(4), ID1=b(5), ORG=b(7), HWM=b(9), BGR=b(12),
    DFM=b(14), TRI=b(15),
    //Reg 04h Resize control
    RSZ0=b(0), RSZ1=b(1), RCH0=b(4), RCH1=b(5), RCV0=b(8), RCV1=b(9),
    //Reg 07h Display control 1
    D0=b(0), D1=b(1), CL=b(3), DTE=b(4), GON=b(5), BASEE=b(8),
    PTDE0=b(12), PTDE1=b(13),
    //Reg 08h Display Control 2
    BP0=b(0), BP1=b(1), BP2=b(2), BP3=b(3),
    FP0=b(8), FP1=b(9), FP2=b(10), FP3=b(11),
    //Reg 09h Display Control 3
    ISC0=b(0), ISC1=b(1), ISC2=b(2), ISC3=b(3),
    PTG0=b(4), PTG1=b(5),
    PTS0=b(8), PTS1=b(9), PTS2=b(10),
    //Reg 0Ah Display Control 4
    FMI0=b(0), FMI1=b(1), FMI2=b(3), FMARKOE=b(3),
    //Reg 0Ch RGB Display Interface Control 1
    ENC2=b(14), ENC1=b(13), ENC0=b(12), RM=b(8), DM1=b(5), DM0=b(4),
    RIM1=b(1), RIM0=b(0),
    //Reg 0Dh Frame Maker Position
    FMP8=b(8), FMP7=b(7), FMP6=b(6), FMP5=b(5), FMP4=b(4), FMP3=b(3),
    FMP2=b(2), FMP1=b(1), FMP0=b(0),
    //Reg 0Fh RGB Display Interface Control 2
    EPL=b(0), DPL=b(1), HSPL=b(3), VSPL=b(4),
    //Reg 10h Power Control 1
    SAP=b(12), BT2=b(10), BT1=b(9), BT0=b(8), APE=b(7),
    AP2=b(6), AP1=b(5), AP0=b(4), DSTB=b(2), SLP=b(1), STB=b(0),
    //Reg 11h Power Control 2
    DC12=b(10), DC11=b(9), DC10=b(8), DC02=b(6), DC01=b(5), DC00=b(4),
    VC2=b(2), VC1=b(1), VC0=b(0),
    //Reg 12h Power Control 3
    VCIRE=b(7), PON=b(4), VRH3=b(3), VRH2=b(2), VRH1=b(1), VRH0=b(0),
    //Reg 13h Power Control 4
    VDV4=b(12), VDV3=b(11), VDV2=b(10), VDV1=b(9), VDV0=b(8),
    //Reg 20h Horizontal GRAM Address Set (0..239)
    //Reg 21h Vertical GRAM Address Set (0...319)
    //Reg 22h Write Data to GRAM
    //Reg 29h Power Control 7
    VCM5=b(5), VCM4=b(4), VCM3=b(3), VCM2=b(2), VCM1=b(1), VCM0=b(0),
    //Reg 2Bh Frame Rate and Color Control
    FRS3=b(3), FRS2=b(2), FRS1=b(1), FRS0=b(0),
    //Reg 30h Gamma Control 1:  GREGVAL(KP1, KP0)
    //Reg 31h Gamma Control 2:  GREGVAL(KP3, KP2)
    //Reg 32h Gamma Control 3:  GREGVAL(KP5, KP4)
    //Reg 35h Gamma Control 4:  GREGVAL(RP1, RP0)
    //Reg 36h Gamma Control 5:  GREGVAL(VRP1, VRP0)
    //Reg 37h Gamma Control 6:  GREGVAL(KN1, KN0)
    //Reg 38h Gamma Control 7:  GREGVAL(KN3, KN2)
    //Reg 39h Gamma Control 8:  GREGVAL(KN5, KN4)
    //Reg 3Ch Gamma Control 9:  GREGVAL(RN1, RN0)
    //Reg 3Dh Gamma Control 10: GREGVAL(VRN1, VRN0)
    //Reg 50h Horizontal Address Start Position (0...239)
    //Reg 51h Horizontal Address End Position (0...239)
    //Reg 52h Vertical Address Start Position (0...319)
    //Reg 53h Vertical Address End Position (0...319)
    //Reg 60h Driver Output Control 2
    GS=b(15), NL5=b(13), NL4=b(12), NL3=b(11), NL2=b(10), NL1=b(9), NL0=b(8),
    SCN5=b(5), SCN4=b(4), SCN3=b(3), SCN2=b(2), SCN1=b(1), SCN0=b(0),
    //Reg 61h Base Image Display Control
    NDL=b(2), VLE=b(1), REV=b(0),
    //Reg 6Ah Vertical Scroll Control (0...319)
    //Reg 80h Partial Image 1 Display Position  (9 bits) PTDP00..08
    //Reg 81h Partial Image 1 Area (Start Line) (9 bits) PTSA00..08
    //Reg 82h Partial Image 1 Area (End Line) (9 bits) PTEA00..08
    //Reg 83h Partial Image 2 Display Position (9 bits) PTDP10..18
    //Reg 84h Partial Image 2 Area (Start Line) (9 bits) PTSA10..18
    //Reg 85h Partial Image 2 Area (End Line) (9 bits) PTEA10..18
    //Reg 90h Panel Interface Control 1
    DIVI1=b(9), DIVI00=b(8), RTNI3=b(3), RTNI2=b(2), RTNI1=b(1), RTNI0=b(0),
    //Reg 92h Panel Interface Control 2
    NOWI2=b(10), NOWI1=b(9), NOWI0=b(8),
    //Reg 95h Panel Interface Control 4
    DIVE1=b(9), DIVE0=b(8),
    RTNE5=b(5), RTNE4=b(4), RTNE3=b(3), RTNE2=b(2), RTNE1=b(1), RTNE0=b(0),
    //Reg A1h OTP VCM Programming Control OTP_PGM_EN + 6 bits (VCM_OTP0..5)
    OTP_PGM_EN=b(11),
    //Reg A2h OTP VCM Status and Enable
    PGM_CNT1=b(15), PGM_CNT0=b(14),
    VCM_D5=b(13), VCM_D4=b(12), VCM_D3=b(11), VCM_D2=b(10),
    VCM_D1=b(9), VCM_D0=b(8),
    VCM_EN=b(0),
    //Reg A5h OTP Programming ID Key (16 bits)
}; //enum

typedef struct __attribute__((packed))
{
    volatile uint16_t IDX;//Index, select register to write
    uint8_t padding[131070];
    union
    {
        volatile uint16_t RAM;
        volatile uint32_t TWOPIX_RAM;
    };
}
DisplayMemLayout;

//==============================================================================
//                  Module's static variables
//==============================================================================
//Pointer to the memory mapped display.
static DisplayMemLayout* const DISPLAY = (DisplayMemLayout* const)0x60000000;

//==============================================================================
//                  Module's static functions
//==============================================================================
static inline void _DispResetOn(void)
{
    GPIOE->BRR = GPIO_Pin_1;
}

static inline void _DispResetOff(void)
{
    GPIOE->BSRR = GPIO_Pin_1;
}

static inline void _DispBacklightOn(void)
{
    GPIOD->BSRR = GPIO_Pin_13;
}

static inline void _DispBacklightOff(void)
{
    GPIOD->BRR = GPIO_Pin_13;
}

static inline void _writeIdx(uint8_t reg)
{
    DISPLAY->IDX = reg;
}

static inline void _writeRam(uint16_t data)
{
    DISPLAY->RAM = data;
}

static inline uint16_t _readRam()
{
    return DISPLAY->RAM;
}

static inline void _writeReg(uint8_t reg, uint16_t data)
{
    DISPLAY->IDX = reg;
    DISPLAY->RAM = data;
}

static inline uint16_t _readReg(uint8_t reg)
{
    DISPLAY->IDX = reg;
    return DISPLAY->RAM;
}

static inline uint16_t _min(uint16_t a, uint16_t b)
{
    return a > b ? b : a;
}

static inline uint16_t _max(uint16_t a, uint16_t b)
{
    return a > b ? a : b;
}

static inline int _abs(int a)
{
    return (a < 0) ? -a : a;
}

static inline void _setCursor(LCDPoint p)
{
#ifdef LCD_ORIENTATION_VERTICAL
    assert_param(p.x < 240);
    assert_param(p.y < 320);
    _writeReg(0x20, p.x);
    _writeReg(0x21, p.y);
#elif defined LCD_ORIENTATION_HORIZONTAL
    assert_param(p.x < 320);
    assert_param(p.y < 240);
    _writeReg(0x20, p.y);
    _writeReg(0x21, 319 - p.x);
#endif
}

static inline void _LimitedWindow(LCDPoint p1, LCDPoint p2)
{
    assert_param(p2.x >= p1.x);
    assert_param(p2.y >= p1.y);
#ifdef LCD_ORIENTATION_VERTICAL
    assert_param(p2.x < 240);
    assert_param(p2.y < 320);
    _writeReg(0x03, BGR | ID1 | ID0 | AM);//GRAM increment up-to-down first, then left-to-right
    _writeReg(0x50, p1.x);
    _writeReg(0x51, p2.x);
    _writeReg(0x52, p1.y);
    _writeReg(0x53, p2.y);
    _setCursor(p1);
#elif defined LCD_ORIENTATION_HORIZONTAL
    assert_param(p2.x < 320);
    assert_param(p2.y < 240);
    _writeReg(0x03, (BGR | ID0));//GRAM increment left-to-right first, then up to down
    _writeReg(0x50, p1.y);
    _writeReg(0x51, p2.y);
    _writeReg(0x52, 319 - p2.x);
    _writeReg(0x53, 319 - p1.x);
    _setCursor(p1);
#endif
}

static void _BresenhamLine(LCDPoint a, LCDPoint b, LCDColor c)
{
    //Bresenham's algorithm
    const short dx = b.x - a.x;
    const short dy = b.y - a.y;
    const short adx = _abs(dx);
    const short ady = _abs(dy);

#ifdef LCD_ORIENTATION_VERTICAL
    assert_param(a.x < 240);
    assert_param(a.y < 320);
    assert_param(b.x < 240);
    assert_param(b.y < 320);
#elif defined LCD_ORIENTATION_HORIZONTAL
    assert_param(a.x < 320);
    assert_param(a.y < 240);
    assert_param(b.x < 320);
    assert_param(b.y < 240);
#endif

    if (adx > ady)
    {
        short yincr = dy >= 0 ? 1 : -1;
        short d = 2 * ady - adx;
        short v = 2 * (ady - adx);
        short w = 2 * ady;
        short y = a.y;
        short x;
        if (dx > 0)
        {
            for (x = a.x; x <= b.x; x++)
            {
                LCD_SetPixel(LCD_MakePoint(x, y), c);
                if (d > 0)
                {
                    y += yincr;
                    d += v;
                }
                else
                    d += w;
            }
        }
        else
        {
            for (x = a.x; x >= b.x; x--)
            {
                LCD_SetPixel(LCD_MakePoint(x, y), c);
                if(d > 0)
                {
                    y += yincr;
                    d += v;
                }
                else
                    d += w;
            }
        }
    }
    else
    {
        short xincr = dx >= 0 ? 1 : -1;
        short d = 2 * adx - ady;
        short v = 2 * (adx - ady);
        short w = 2 * adx;
        short x = a.x;
        short y;
        if(dy > 0)
        {
            for (y = a.y; y <= b.y; y++)
            {
                LCD_SetPixel(LCD_MakePoint(x, y), c);
                if(d > 0)
                {
                    x += xincr;
                    d += v;
                }
                else
                    d += w;
            }
        }
        else
        {
            for (y = a.y; y >= b.y; y--)
            {
                LCD_SetPixel(LCD_MakePoint(x, y), c);
                if (d > 0)
                {
                    x += xincr;
                    d += v;
                }
                else
                    d += w;
            }
        }
    }
} // _BresenhamLine

//==============================================================================
//                  Module's public API
//==============================================================================
void LCD_DelayMs(uint32_t ms)
{
    uint32_t i;
#ifdef SYSCLK_FREQ_72MHz
    register const uint32_t count = 12000;
#elif SYSCLK_FREQ_56MHz
    register const uint32_t count = 11200;
#elif SYSCLK_FREQ_48MHz
    register const uint32_t count = 9600;
#elif SYSCLK_FREQ_36MHz
    register const uint32_t count = 7200;
#elif SYSCLK_FREQ_24MHz
    register const uint32_t count = 6000;
#else
    register const uint32_t count = 2016;
#endif
    for (i = 0; i < ms; i++)
    {
        // This delay has been calibrated to take 1 millisecond
        // It is written in assembler to be independent on compiler optimizations
        asm volatile ("           mov   r1, #0     \n"
                      "1:         cmp   r1, %0     \n"
                      "           itt   lo         \n"
                      "           addlo r1, r1, #1 \n"
                      "           blo   1b  \n"::"r"(count):"r1");
    }
}

LCDPoint LCD_MakePoint(int x, int y)
{
    if (x < 0) x = 0;
    if (y < 0) y = 0;
    LCDPoint pt = {x, y};
    return pt;
}

void LCD_Init(void)
{
    volatile uint32_t* pBCR1 = &FSMC_Bank1->BTCR[0];
    volatile uint32_t* pBTR1 = &FSMC_Bank1->BTCR[1];
    GPIO_InitTypeDef gpioStruct;

    RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOD | RCC_APB2Periph_GPIOE, ENABLE);
    GPIO_StructInit(&gpioStruct);

    //Databus + wr, rd, cs, rs
    gpioStruct.GPIO_Speed = GPIO_Speed_50MHz;
    gpioStruct.GPIO_Pin = GPIO_Pin_0 | GPIO_Pin_1 | GPIO_Pin_8 | GPIO_Pin_9
                          | GPIO_Pin_10 | GPIO_Pin_14 | GPIO_Pin_15
                          | GPIO_Pin_4 | GPIO_Pin_5| GPIO_Pin_7 | GPIO_Pin_11;
    gpioStruct.GPIO_Mode = GPIO_Mode_AF_PP;
    GPIO_Init(GPIOD, &gpioStruct);
    gpioStruct.GPIO_Pin = GPIO_Pin_7 | GPIO_Pin_8 | GPIO_Pin_9 | GPIO_Pin_10
                          | GPIO_Pin_11 | GPIO_Pin_12 | GPIO_Pin_13 | GPIO_Pin_14
                          | GPIO_Pin_15;
    gpioStruct.GPIO_Mode = GPIO_Mode_AF_PP;
    GPIO_Init(GPIOE, &gpioStruct);
    //Backlight
    gpioStruct.GPIO_Pin = GPIO_Pin_13;
    gpioStruct.GPIO_Mode = GPIO_Mode_Out_PP;
    GPIO_Init(GPIOD, &gpioStruct);
    //Reset
    gpioStruct.GPIO_Pin = GPIO_Pin_1;
    gpioStruct.GPIO_Mode = GPIO_Mode_Out_PP;
    GPIO_Init(GPIOE, &gpioStruct);

    //Configure FSMC interface to LCD
    RCC->AHBENR |= RCC_AHBENR_FSMCEN;

    *pBCR1 = FSMC_BCR1_MBKEN  | //memory bank enable
             FSMC_BCR1_MWID_0 | //memory data width: 16
             FSMC_BCR1_WREN;    //Write Operation Enable

    *pBTR1 = 0 << 0 |  // Address setup phase duration
             1 << 4 |  // Address-hold phase duration
             1 << 8 |  // Data-phase duration
             0 << 16 | // Bus turn-around duration
             1 << 20 | // Clock divide ratio
             0 << 24 | // Data latency
             1 << 28;  // Access mode

    //Power up
    LCD_DelayMs(100);
    _DispResetOn();
    LCD_DelayMs(1);
    _DispResetOff();
    LCD_DelayMs(10);

    //=======================================================================
    // The voodoo begins here. Don't touch the code unless you fully realize
    // what you are doing: the display may be seriously damaged.
    //=======================================================================
    _writeReg(0x00, 0x0001); // Start Oscillation
    _writeReg(0x01, SS); // SS=1, SM=0
    _writeReg(0x02, BC0 | EOR); // BC0=1 EOR=1

    _writeReg(0x03, BGR | ID1 | ID0 | AM); //Entry mode
    _writeReg(0x04, 0x0000); // Resize register
    _writeReg(0x08, 0x0202); // Front and Back Porch (must be >= 2 lines)
    _writeReg(0x09, 0x0000); // Normal scan, Scan cycle=0
    _writeReg(0x0A, 0x0000); // Output of FMARK signal is disabled
    _writeReg(0x0C, RIM0);   // System interface, 16 bits, internal clock
    _writeReg(0x0D, 0x0000); // Frame marker Position
    _writeReg(0x0F, 0x0000); // System interface polarity (CLK, EN)

    //Power-up sequence
    _writeReg(0x10, 0x0000);
    _writeReg(0x11, VC0|VC1|VC2); //Vci1 = 1.0 * Vci
    _writeReg(0x12, 0x0000);      //External reference
    _writeReg(0x13, 0x0000);      //VCOM = 0.70 * VREG1OUT
    LCD_DelayMs(200);                //Let the voltage stabilize
    _writeReg(0x10, SAP | BT2 | BT1 | APE | AP0); //Enable power supply circuits
    _writeReg(0x11, VC0 | DC01 | DC11); //Step-up circuits parameters
    LCD_DelayMs(50);                 // Let it stabilize
    _writeReg(0x12, PON | VRH3 | VRH2); //Set grayscale level
    LCD_DelayMs(50);                 // Let it stabilize
    _writeReg(0x13, VDV4 | VDV1 | VDV0); //Set Vcom voltage
    _writeReg(0x29, VCM4 | VCM3 | VCM2); // Set VcomH voltage
    _writeReg(0x2B, 0x000D); //Set frame rate 93Hz

    //Gamma
    //Setting all values for the linear curve, except RN1 (to increase contrast
    //in the darkest area)
    _writeReg(0x30, GREGVAL(3,  3));  //KP1[3] KP0[3]   Fine adjustment positive: grayscale 8, grayscale 1
    _writeReg(0x31, GREGVAL(3,  3));  //KP3[3] KP2[3]   Fine adjustment positive: grayscale 43, grayscale 20
    _writeReg(0x32, GREGVAL(3,  3));  //KP5[3] KP4[3]   Fine adjustment positive: grayscale 62, grayscale 55
    _writeReg(0x35, GREGVAL(1,  1));  //RP1[3] RP0[3]   Gradient adjustment variable resistors, positive
    _writeReg(0x36, GREGVAL(15, 15)); //VRP1[5] VRP0[5] Amplitude adjustment variable resistors, positive
    _writeReg(0x37, GREGVAL(3,  3));  //KN1[3] KN0[3]   Fine adjustment negative: grayscale 8, grayscale 1
    _writeReg(0x38, GREGVAL(3,  3));  //KN3[3] KN2[3]   Fine adjustment negative: grayscale 43, grayscale 20
    _writeReg(0x39, GREGVAL(3,  3));  //KN5[3] KN4[3]   Fine adjustment negative: grayscale 62, grayscale 55
    _writeReg(0x3C, GREGVAL(7,  1));  //RN1[3] RN0[3]   Gradient adjustment variable resistors, negative
    _writeReg(0x3D, GREGVAL(15, 15)); //VRN1[5] VRN0[5] Amplitude adjustment variable resistors, negative

    //GRAM
    _writeReg(0x50, 0x0000); // Horizontal GRAM Start Address
    _writeReg(0x51, 0x00EF); // Horizontal GRAM End Address
    _writeReg(0x52, 0x0000); // Vertical GRAM Start Address
    _writeReg(0x53, 0x013F); // Vertical GRAM Start Address
    _writeReg(0x60, 0xA700); // Gate Scan Line
    _writeReg(0x61, 0x0001); // NDL,VLE, REV
    _writeReg(0x6A, 0x0000); // Scrolling line

    //Partial Display
    _writeReg(0x80, 0x0000);
    _writeReg(0x81, 0x0000);
    _writeReg(0x82, 0x0000);
    _writeReg(0x83, 0x0000);
    _writeReg(0x84, 0x0000);
    _writeReg(0x85, 0x0000);

    //Panel Control
    _writeReg(0x90, 0x0010);
    _writeReg(0x92, 0x0000);
    _writeReg(0x93, 0x0003);
    _writeReg(0x95, 0x0110);
    _writeReg(0x97, 0x0000);
    _writeReg(0x98, 0x0000);

    // Display ON
    _writeReg(0x07, D0 | D1 |DTE | GON | BASEE);

    //Fill display
    LCD_FillAll(LCD_BLACK);

    //Turn on backlight
    _DispBacklightOn();
}

void LCD_FillAll(LCDColor c)
{
    LCD_FillRect(LCD_MakePoint(0, 0), LCD_MakePoint(LCD_GetWidth() - 1, LCD_GetHeight() - 1), LCD_BLACK);
}

LCDColor LCD_MakeRGB(uint8_t r, uint8_t g, uint8_t b)
{
    return LCD_RGB(r, g, b);
}

void LCD_FillRect(LCDPoint p1, LCDPoint p2, LCDColor color)
{
    int32_t numPixels;
    int32_t fastPixels;
    int32_t i;
    uint32_t twoPixColor;
    uint16_t tmp;

    if (p1.x >= LCD_GetWidth()) p1.x = LCD_GetWidth() - 1;
    if (p2.x >= LCD_GetWidth()) p2.x = LCD_GetWidth() - 1;
    if (p1.y >= LCD_GetHeight()) p1.y = LCD_GetHeight() - 1;
    if (p2.y >= LCD_GetHeight()) p2.y = LCD_GetHeight() - 1;

    if (p1.x > p2.x)
    {
        tmp = p1.x;
        p1.x = p2.x;
        p2.x = tmp;
    }
    if (p1.y > p2.y)
    {
        tmp = p1.y;
        p1.y = p2.y;
        p2.y = tmp;
    }
    _LimitedWindow(p1, p2);
    _writeIdx(0x22); //Write to GRAM
    numPixels = (p2.x - p1.x + 1) * (p2.y - p1.y + 1);
    fastPixels = numPixels / 2;
    twoPixColor = ((uint32_t)color) | (((uint32_t)color) << 16);
    for (i = 0; i < fastPixels; ++i)
        DISPLAY->TWOPIX_RAM = twoPixColor;
    if (numPixels & 0x1)
        _writeRam(color);
}

void LCD_Rectangle(LCDPoint a, LCDPoint b, LCDColor c)
{
    LCD_Line(a, LCD_MakePoint(b.x, a.y), c);
    LCD_Line(LCD_MakePoint(b.x, a.y), b, c);
    LCD_Line(b, LCD_MakePoint(a.x, b.y), c);
    LCD_Line(LCD_MakePoint(a.x, b.y), a, c);
}

void LCD_TurnOn(void)
{
    _writeReg(0x10, 0); //Exit standby mode
    _writeReg(0x10, SAP | BT2 | BT1 | APE | AP0); //Enable power supply circuits
    _DispBacklightOn(); //Backlight on
}

void LCD_TurnOff(void)
{
    _DispBacklightOff();  //Backlight off
    _writeReg(0x10, STB); //Standby (Only GRAM continues operation).
}

void LCD_SetPixel(LCDPoint p, LCDColor color)
{
    if (p.x >= LCD_GetWidth() || p.y >= LCD_GetHeight())
        return;
    _setCursor(p);
    _writeIdx(0x22); //Write to GRAM
    _writeRam(color);
}

void LCD_Line(LCDPoint a, LCDPoint b, LCDColor color)
{
    int numPixels;
    int fastPixels;
    int i;
    uint32_t twoPixColor;

    if (a.x >= LCD_GetWidth())
        a.x = LCD_GetWidth() - 1;
    if (b.x >= LCD_GetWidth())
        b.x = LCD_GetWidth() - 1;
    if (a.y >= LCD_GetHeight())
        a.y = LCD_GetHeight() - 1;
    if (b.y >= LCD_GetHeight())
        b.y = LCD_GetHeight() - 1;
    //Horizontal line speed optimization
    if(a.y == b.y)
    {
        _LimitedWindow(LCD_MakePoint(_min(a.x, b.x), a.y),
                       LCD_MakePoint(_max(a.x, b.x), a.y));
        _writeIdx(0x22); //Write to GRAM
        numPixels = _abs(a.x - b.x);
        fastPixels = numPixels / 2;
        twoPixColor = ((uint32_t)color) | (((uint32_t)color) << 16);
        for (i = 0; i <= fastPixels; ++i)
            DISPLAY->TWOPIX_RAM = twoPixColor;
        if (numPixels & 0x1) _writeRam(color);
        return;
    }

    //Vertical line speed optimization
    if(a.x == b.x)
    {
        _LimitedWindow(LCD_MakePoint(a.x, _min(a.y, b.y)),
                       LCD_MakePoint(a.x, _max(a.y, b.y)));
        _writeIdx(0x22); //Write to GRAM
        numPixels = _abs(a.y - b.y);
        fastPixels = numPixels / 2;
        twoPixColor = ((uint32_t)color) | (((uint32_t)color) << 16);
        for (i = 0; i <= fastPixels; ++i)
            DISPLAY->TWOPIX_RAM = twoPixColor;
        if (numPixels & 0x1)
            _writeRam(color);
        return;
    }
    //For non-horizontnal and non-vertical line use slow
    //programmatic way to draw it pixel-by-pixel
    _BresenhamLine(a, b, color);
} //LCD_Line

uint16_t LCD_GetWidth(void)
{
#ifdef LCD_ORIENTATION_HORIZONTAL
    return 320;
#else
    return 240;
#endif
}

uint16_t LCD_GetHeight(void)
{
#ifdef LCD_ORIENTATION_HORIZONTAL
    return 240;
#else
    return 320;
#endif
}

void LCD_Circle(LCDPoint center, int16_t r, LCDColor color)
{
    int16_t cx = 0;
    int16_t cy = r;
    int16_t df = 1 - r;
    int16_t d_e = 3;
    int16_t d_se = -2 * r + 5;
    if(r == 0)
    {
        return;
    }

    do
    {
        LCD_SetPixel(LCD_MakePoint(center.x + cx, center.y - cy), color);
        LCD_SetPixel(LCD_MakePoint(center.x + cx, center.y + cy), color);
        LCD_SetPixel(LCD_MakePoint(center.x - cx, center.y - cy), color);
        LCD_SetPixel(LCD_MakePoint(center.x - cx, center.y + cy), color);
        LCD_SetPixel(LCD_MakePoint(center.x + cy, center.y + cx), color);
        LCD_SetPixel(LCD_MakePoint(center.x + cy, center.y - cx), color);
        LCD_SetPixel(LCD_MakePoint(center.x - cy, center.y + cx), color);
        LCD_SetPixel(LCD_MakePoint(center.x - cy, center.y - cx), color);

        if (df < 0)
        {
            df += d_e;
            d_e += 2;
            d_se += 2;
        }
        else
        {
            df += d_se;
            d_e += 2;
            d_se += 4;
            cy--;
        }
        cx++;
    }
    while (cx <= cy);
}

void LCD_FillCircle(LCDPoint center, int16_t r, LCDColor color)
{
    int16_t cx = 0;
    int16_t cy = r;
    int16_t df = 1 - r;
    int16_t d_e = 3;
    int16_t d_se = -2 * r + 5;
    if(r == 0)
    {
        return;
    }

    do
    {
        if (df >= 0)
        {
            LCD_Line(LCD_MakePoint(center.x - cx, center.y + cy), LCD_MakePoint(center.x + cx, center.y + cy), color);
            LCD_Line(LCD_MakePoint(center.x - cx, center.y - cy), LCD_MakePoint(center.x + cx, center.y - cy), color);
        }
        if (cx != cy)
        {
            LCD_Line(LCD_MakePoint(center.x - cy, center.y + cx), LCD_MakePoint(center.x + cy, center.y + cx), color);
            LCD_Line(LCD_MakePoint(center.x - cy, center.y - cx), LCD_MakePoint(center.x + cy, center.y - cx), color);
        }
        if (df < 0)
        {
            df += d_e;
            d_e += 2;
            d_se += 2;
        }
        else
        {
            df += d_se;
            d_e += 2;
            d_se += 4;
            cy--;
        }
        cx++;
    }
    while (cx <= cy);
}


//Drawing bitmap file based on libnsbmp

#define BYTES_PER_PIXEL 4
#define TRANSPARENT_COLOR 0xFFFFFFFF

void *bitmap_create(int width, int height, unsigned int state)
{
    (void) state;  /* unused */
    return (void*)0x2000F000; //calloc(width * height, BYTES_PER_PIXEL);
}


unsigned char *bitmap_get_buffer(void *bitmap)
{
    return (unsigned char*)bitmap;
}


size_t bitmap_get_bpp(void *bitmap)
{
    (void) bitmap;  /* unused */
    return BYTES_PER_PIXEL;
}


void bitmap_destroy(void *bitmap)
{
    //free(bitmap);
}

static LCDPoint _bmpOrigin;

void bitmap_putcolor(unsigned int color32, unsigned int x, unsigned int y)
{
    //printf("Color: 0x%08x, x: %d, y: %d\n", color32, x, y);
    uint16_t r, g, b;
    r = (uint16_t)(color32 & 0xFF);
    g = (uint16_t)((color32 & 0xFF00) >> 8);
    b = (uint16_t)((color32 & 0xFF0000) >> 16);
    LCD_SetPixel(LCD_MakePoint(_bmpOrigin.x + x, _bmpOrigin.y + y), LCD_RGB(r, g, b));
}

static bmp_bitmap_callback_vt bitmap_callbacks =
{
    bitmap_create,
    bitmap_destroy,
    bitmap_get_buffer,
    bitmap_get_bpp,
    bitmap_putcolor
};

void LCD_DrawBitmap(LCDPoint origin, const uint8_t *bmpData, uint32_t bmpDataSize)
{
    bmp_image bmp;
    bmp_result code;
    _bmpOrigin = origin;

    bmp_create(&bmp, &bitmap_callbacks);
    code = bmp_analyse(&bmp, bmpDataSize, bmpData);
    if(code != BMP_OK)
    {
        //printf("bmp_analyse failed, %d", code);
        goto cleanup;
    }
    code = bmp_decode(&bmp);
    if(code != BMP_OK)
    {
        //printf("bmp_decode failed, %d", code);
        if(code != BMP_INSUFFICIENT_DATA)
        {
            goto cleanup;
        }
    }
    //printf("# width                %d \n", (int)bmp.width);
    //printf("# height               %d \n", (int)bmp.height);
cleanup:
    bmp_finalise(&bmp);
}
