#include "touch.h"

//This module requires STM32F10x Std Peripheral Library
#include "stm32f10x.h"

/*
 Calibration points of my LCD:
x:350 y:3780                       x:3575 y:3725
                 x:2000 y:2000
x:350 y:255                        x:3170 y:280
*/

//Calibration values for raw TSC2046 data. May be specific for particular display.
#define XMAX 3450
#define XMIN 350
static const uint32_t XBW = XMAX - XMIN; //x bandwidth = 0...3400
#define YMAX 3750
#define YMIN 280
static const uint32_t YBW = YMAX - YMIN; //y bandwidth = 0...3495

static void _def_event_callback(TOUCH_EVENTTYPE evType, LCDPoint point)
{
    (void)evType;
    (void)point;
}

static TOUCH_EVENTCALLBACK _callback = _def_event_callback;

static uint32_t TOUCH_rawx = 0;
static uint32_t TOUCH_rawy = 0;

unsigned int TOUCH_GetRawX(void)
{
    return TOUCH_rawx;
}

unsigned int TOUCH_GetRawY(void)
{
    return TOUCH_rawy;
}

//HW specific functions
static inline int _spi1_touchint_value(void)
{
    return (GPIOB->IDR & GPIO_Pin_6) ? 1 : 0;
}

static inline int _spi1_miso_value(void)
{
    return (GPIOA->IDR & GPIO_Pin_6) ? 1 : 0;
}

static inline void _spi1_sck_low(void)
{
    GPIOA->BRR = GPIO_Pin_5;
}

static inline void _spi1_sck_high(void)
{
    GPIOA->BSRR = GPIO_Pin_5;
}

static inline void _spi1_mosi_low(void)
{
    GPIOA->BRR = GPIO_Pin_7;
}

static inline void _spi1_mosi_high(void)
{
    GPIOA->BSRR = GPIO_Pin_7;
}

static inline void _delayUs(uint32_t us)
{
    //Calibrated for 72 MHz and code in flash
    asm volatile("           mov   r2, #14    \n"
                 "           mov   r1, #0     \n"
                 "2:         cmp   r1, r2     \n"
                 "           itt   lo         \n"
                 "           addlo r1, r1, #1 \n"
                 "           blo   2b         \n"
                 "           mov   r1, #8     \n"
                 "           mul   r2, %0, r1 \n"
                 "           mov   r1, #0     \n"
                 "1:         nop              \n"
                 "           cmp   r1, r2     \n"
                 "           itt   lo         \n"
                 "           addlo r1, r1, #1 \n"
                 "           blo   1b  \n"::"r"(us):"r1","r2");
}

static inline void _spi1_ntouchss_high(void)
{
    GPIOB->BSRR = GPIO_Pin_7;
}

static inline void _spi1_ntouchss_low(void)
{
    GPIOB->BRR = GPIO_Pin_7;
}

/**
 Exchange byte over SPI interface.
 @note SPI is implemented by bitbanging because with APB1 clock of 72 MHz
 it is impossible to use hardware SPI1 on speeds below 285 kb/s. TSC2046
 requires speed of 125 kb/s or less.
 @param [in] byteTx Byte to transmit
 @return received byte
*/
static uint8_t SpiExchangeByte(uint8_t byteTx)
{
    uint8_t mask = 0x80;
    uint8_t rxByte = 0;
    while (mask)
    {
        (byteTx & mask) ? _spi1_mosi_high() : _spi1_mosi_low();
        _delayUs(5);
        if (_spi1_miso_value())
            rxByte |= mask;
        _spi1_sck_high();
        _delayUs(5);
        _spi1_sck_low();
        mask >>= 1;
    }
    //_spi1_mosi_low();
    return rxByte;
}


/** Poll touchscreen controller once. \n
    @note This function returns raw data measured by ADC. It is an overhead
    to convert it to coordinates here. It's better to do this after filtering
    several measurements.
    @param [in] coord the structure that will be filled with raw ADC data
    @return true if successful, false if touch is released.
*/
static int Poll2046Once(LCDPoint* pCoord)
{
    //16-bits, DFR mode, measure Y, X.
    static const uint8_t txBytes[5] = {0x92, 0x00, 0xD2, 0x00, 0x00};
    uint8_t rxBytes[5];

    uint32_t i;

    if (0 == pCoord)
        return 0;

    if (_spi1_touchint_value())
        return 0;

    //Disable context switching. It is enough.
    //PauseKernelLock kLock;

    //Chip select
    _spi1_ntouchss_low();
    _delayUs(5);

    //Exchange bytes
    for (i = 0; i < 5; ++i)
    {
        rxBytes[i] = SpiExchangeByte(txBytes[i]);
    }
    //Deselect
    _spi1_ntouchss_high();
    _delayUs(2);

    //Touch has been released before ADC conversion is finished:
    //the result can be invalid.
    if (_spi1_touchint_value())
        return 0;

    //Now extract raw ADC values
    uint16_t rawY =  (((uint16_t)rxBytes[1]) << 5) | (((uint16_t)rxBytes[2]) >> 3);
    uint16_t rawX =  (((uint16_t)rxBytes[3]) << 5) | (((uint16_t)rxBytes[4]) >> 3);
    pCoord->x = rawX;
    pCoord->y = rawY;
    return 1;
}

int TOUCH_IsPressed(void)
{
    return !_spi1_touchint_value();
}


/** Poll touchscreen controller, filter data and calculate real coordinates.
    @param [in] coord the structure that will be filled with raw ADC data
    @return true if successful, false if touch is released.
*/
int TOUCH_Poll(LCDPoint* pCoord)
{
    LCDPoint polls[5];
    uint16_t max = 0;
    uint16_t min = 4096;
    int imin = 0;
    int imax = 0;
    uint16_t x = 0;
    uint16_t y = 0;
    uint16_t resX = 0;
    uint16_t resY = 0;

    if (0 == pCoord)
        return 0;

    //Poll controller 5 times
    int i;
    for (i = 0; i < 5; ++i)
    {
        if (0 == Poll2046Once(&polls[i]))
            return 0; //touch has been released
        _delayUs(500);
        //extern void prbytes(int i, uint8_t* data);
        //prbytes(i, rxBytes);
    }

    //Now find max and min x to exclude
    for (i = 0; i < 5; ++i)
    {
        if (polls[i].x < min) { min = polls[i].x; imin = i; }
        if (polls[i].x > max) { max = polls[i].x; imax = i; }
    }
    if (imin == imax)
    {//no need to calculate median
        x = polls[imin].x;
    }
    else
    {
        x = 0;
        for (i = 0; i < 5; ++i)
        {
            if (i == imin || i == imax) continue; //Skip
            x += polls[i].x;
        }
        x /= 3; //Median value of x
    }

    //Repeat the same for y
    max = 0;
    min = 4096;
    for (i = 0; i < 5; ++i)
    {
        if (polls[i].y < min) { min = polls[i].y; imin = i; }
        if (polls[i].y > max) { max = polls[i].y; imax = i; }
    }

    if (imin == imax)
    {//no need to calculate median
        y = polls[imin].y;
    }
    else
    {
        y = 0;
        for (i = 0; i < 5; ++i)
        {
            if (i == imin || i == imax) continue;
            y += polls[i].y;
        }
        y /= 3; //Median value of y
    }

    TOUCH_rawx = x; //Store for calibration purposes
    TOUCH_rawy = y;

    if (x > XMAX) x = XMAX;
    else if (x < XMIN) x = XMIN;
    x -= XMIN;
    if (y > YMAX) y = YMAX;
    else if (y < YMIN) y = YMIN;
    y -= YMIN;

    //Now convert to real coordinates on screen
#if defined(LCD_ORIENTATION_VERTICAL)
    resX = 240 - (((uint32_t)y) * 240) / YBW;
    resY = 320 - (((uint32_t)x) * 320) / XBW;
#elif defined(LCD_ORIENTATION_HORIZONTAL)
    resX = (((uint32_t)x) * 320) / XBW;
    resY = 240 - (((uint32_t)y) * 240) / YBW;
#else
    #error Not implemented
#endif

    if (resX >= LCD_GetWidth())
        resX = LCD_GetWidth() - 1;
    if (resY >= LCD_GetHeight())
        resY = LCD_GetHeight() - 1;
    pCoord->x = resX;
    pCoord->y = resY;
    return 1;
}


void TOUCH_EventsThread(void)
{
    int tPrev = 0;
    LCDPoint pOld = {0};
    LCDPoint p = {0};

    for(;;)
    {
        //Thread::sleep(20); //Check for events 50 times a second

        //PollButtons();

        //Check touchscreen
        if( TOUCH_Poll(&p) )
        {//Someone is touching the screen
            //Did the touch point differ that much from the previous?
            //if(abs(pOld.x()-p.x()) > 3 || abs(pOld.y()-p.y()) > 3 || !tPrev)
            if( !tPrev || pOld.x != p.x  || pOld.y != p.y)
            {
                pOld = p;
                if(tPrev == 0)
                    _callback(TOUCH_EVENT_TOUCHDOWN, pOld);
                else
                    _callback(TOUCH_EVENT_TOUCHMOVE, pOld);
            }
            tPrev = 1;
        }
        else
        {
            //No, no one is touching the screen
            if (tPrev)
                _callback(TOUCH_EVENT_TOUCHUP, pOld);
            tPrev = 0;
        }
    }// for(;;)
}

void TOUCH_SetCallback(TOUCH_EVENTCALLBACK cb)
{
    if (0 != cb)
        _callback = cb;
}


void TOUCH_Init(void)
{
    GPIO_InitTypeDef gpioStruct;

    RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOA | RCC_APB2Periph_GPIOB, ENABLE);
    GPIO_StructInit(&gpioStruct);

    gpioStruct.GPIO_Speed = GPIO_Speed_50MHz;
    gpioStruct.GPIO_Pin = GPIO_Pin_7 | GPIO_Pin_5; //mosi and sck
    gpioStruct.GPIO_Mode = GPIO_Mode_Out_PP;
    GPIO_Init(GPIOA, &gpioStruct);
    GPIOA->BRR = GPIO_Pin_5; //sck low

    gpioStruct.GPIO_Pin = GPIO_Pin_7; //ntouchss
    gpioStruct.GPIO_Mode = GPIO_Mode_Out_PP;
    GPIO_Init(GPIOB, &gpioStruct);
    GPIOB->BSRR = GPIO_Pin_7; //ntouchss high

    gpioStruct.GPIO_Pin = GPIO_Pin_6; //miso
    gpioStruct.GPIO_Mode = GPIO_Mode_IPU;
    GPIO_Init(GPIOA, &gpioStruct);

    gpioStruct.GPIO_Pin = GPIO_Pin_6 | GPIO_Pin_15; //touchint, button
    gpioStruct.GPIO_Mode = GPIO_Mode_IPU;
    GPIO_Init(GPIOB, &gpioStruct);

    //LCD_DelayMs(10);
}

