//==============================================
// EU3EU antenna analyzer project
// Main file
//==============================================

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <math.h>
#include "stm32f10x.h"
#include "font.h"
#include "LCD.h"
#include "swr1.h" //VSWR scan mode window background bitmap
#include "touch.h"
#include "dds.h"
#include "ad8302.h"

/*********************************************************************
 *
 *  main()
 *
 *********************************************************************/
asm(".global _printf_float");

static volatile uint32_t _msCounter = 0;

extern "C"
void SysTick_Handler()
{
    ++_msCounter;
}

extern "C"
uint32_t GetTickCount()
{
    return _msCounter;
}

extern "C"
void Sleep(uint32_t nms)
{
    uint32_t tStart = GetTickCount();
    while(GetTickCount() - tStart < nms);
}

void ShowF()
{
    char str[50];
    sprintf(str, "F: %u kHz     ", (unsigned int)(DDS_GetFreq() / 1000));
    FONT_Write(FONT_FRANBIG, LCD_RED, LCD_BLACK, 0, 0, str);
}

/*
    AD8302 is connected to measurement bridge as:

     Signal in  ----------------------
     from DDS     |                  |
                  |------ OFSB       |------ OFSA
                Rmeas               Rmeas
                  |-------INPB       |-------INPA
                  |                  |
                Radd                Radd
                  |                  |
                [Antenna]           Rload
                  |                  |
                 ___      GND       ___
*/

static const float Rmeas = 10.0;
static const float RmeasAdd = 10.0;
static const float Rload = 50.0;
static const float Rtotal = RmeasAdd + Rmeas + Rload;
static const float Z0 = 50.0;

//Display measured data
void ShowADC(UADC in)
{
    float _ratio = AD8302_GetRatio(in);
    float _phase = AD8302_GetPhaseRadians(in);
    char str[50];
    sprintf(str, "%.2f dB %.1f deg         ", AD8302_GetRatioDb(in), AD8302_GetPhaseDegrees(in));
    FONT_Write(FONT_FRANBIG, LCD_WHITE, LCD_BLACK, 0, 30, str);

    float XX = sinf(_phase) * Rtotal * _ratio;
    float RR = (cosf(_phase) * Rtotal * _ratio) - (Rmeas + RmeasAdd);
    if(RR < 5.0)
    {
        RR = 0.0;
    }
    sprintf(str, "R: %.1f  X: %.1f         ", RR, XX);
    FONT_Write(FONT_FRANBIG, LCD_GREEN, LCD_BLACK, 0, 60, str);

    float ro = sqrtf((powf((RR - Z0), 2) + powf(XX, 2)) / (powf((RR + Z0), 2) + powf(XX, 2)));
    if(ro > .999)
    {
        ro = 0.999;
    }
    float VSWR = (1.0 + ro) / (1.0 - ro);
    sprintf(str, "VSWR: %.1f             ", VSWR);
    FONT_Write(FONT_FRANBIG, LCD_CYAN, LCD_BLACK, 0, 90, str);

    //calculate equivalent capacity and inductance (if XX is big enough)
    if(XX > 5.0)
    {
        float Cpf = 1e12 / (2.0 * 3.1415926 * DDS_GetFreq() * XX);
        if(Cpf > 10000.)
        {
            Cpf = 10000.;
        }
        float Luh = 1e6 * XX / (2.0 * 3.1415926 * DDS_GetFreq());
        sprintf(str, "%.0f pF %.1f uH           ", Cpf, Luh);
        FONT_Write(FONT_FRANBIG, LCD_YELLOW, LCD_BLACK, 0, 120, str);
    }
    else
    {
        FONT_ClearLine(FONT_FRANBIG, LCD_BLACK, 120);
    }
}

uint32_t MeasurementFreq = 10000000;
//Handle touch screen in measurement mode
bool Touch()
{
    LCDPoint pt;
    uint32_t step = 100000;
    if(TOUCH_Poll(&pt))
    {
        if (pt.y > 120)
            return true; //request to change window

        if (pt.x < 40 || pt.x > 280)
        {
            step = 500000;
        }
        else if (pt.x < 80 || pt.x > 240)
        {
            step = 100000;
        }
        else
        {
            step = 10000;
        }
        uint32_t currF = MeasurementFreq;
        if(currF > step && currF % step != 0)
        {
            currF -= (currF % step);
        }
        if(currF < 100000)
        {
            currF = 100000;
        }
        if(pt.x > 160)
        {
            if(currF < 51000000)
            {
                MeasurementFreq = currF + step;
                DDS_SetFreq(MeasurementFreq);
            }
            else
            {
                DDS_SetFreq(currF);
            }
        }
        else
        {
            if(currF > 100000)
            {
                MeasurementFreq = currF - step;
                DDS_SetFreq(MeasurementFreq);
            }
            else
            {
                DDS_SetFreq(currF);
            }
        }
        ShowF();
    }
    return false; //no request to change window
}


//Calculate VSWR
float GetVSWR(UADC in)
{
    float _ratio = AD8302_GetRatio(in);
    float _phase = AD8302_GetPhaseRadians(in);
    float XX = sinf(_phase) * Rtotal * _ratio;
    float RR = (cosf(_phase) * Rtotal * _ratio) - (Rmeas + RmeasAdd);
    if(RR < 5.0)
    {
        RR = 0.0;
    }
    float ro = sqrtf((powf((RR - Z0), 2) + powf(XX, 2)) / (powf((RR + Z0), 2) + powf(XX, 2)));
    if(ro > .999)
    {
        ro = 0.999;
    }
    return (1.0 + ro) / (1.0 - ro);
}

//Calculate offset on graph for given VSWR value
int CalcOffset(float swrvalue)
{
    int offset = 0;
    float v = swrvalue - 1.0;
    if(v <= 1.0)
    {
        //vswr 1.0 ... 2.0, step = 0.01, 100 points
        offset = int(v * 100.);
    }
    else if(v <= 2.0)
    {
        //vswr 2.0 ... 3.0, step = 0.02, 50 points
        offset = 100 + int((v - 1.0) * 50.);
    }
    else
    {
        //vswr 3.0 ... 8.0, step = 0.1
        if(v > 7.0)
        {
            v = 7.0;
        }
        offset = 170 + int((v - 4.0) * 10.);
    }
    return offset;
}

static const int nValFull = 29 * 10 + 1;
static const int nValSmall = 24 * 10 + 1;
static float swrvalues[nValFull];
static uint32_t FMidMHz = 8;

static void DrawFreqSmall()
{
    char str[50];
    sprintf(str, "          F0: %u kHz               ", (unsigned int)(FMidMHz * 1000));
    FONT_Write(FONT_FRAN, LCD_YELLOW, LCD_BLACK, 0, 0, str);
}

//Scan +/- 1.2 MHz bandpass around center frequency FMidMHz with 10 kHz step
//and display graph curve
void ScanSmall(void)
{
    DrawFreqSmall();
    int i;
    //scan VSWR over the small band with 10 kHz step
    FONT_Write(FONT_FRAN, LCD_PURPLE, LCD_BLACK, 180, 0, "Scanning pass 2            ");
    for(i = 0; i < nValSmall; i++)
    {
        uint32_t freq = (FMidMHz * 1000 + i * 10 - 1200) * 1000;
        if (freq < 10000)
        {
            swrvalues[i] = 8.;
            continue;
        }
        DDS_SetFreq(freq);
        UADC adc = AD8302_ReadADC();
        swrvalues[i] = GetVSWR(adc);
        if (TOUCH_IsPressed())
        {
            FONT_Write(FONT_FRAN, LCD_PURPLE, LCD_BLACK, 180, 0, "                           ");
            return;
        }
    }
    FONT_Write(FONT_FRAN, LCD_PURPLE, LCD_BLACK, 180, 0, "                           ");
    int lasty = 0;
    for(i = 0; i < nValSmall; i++)
    {
        int offset = CalcOffset(swrvalues[i]);
        //zero point x = 35 y = 229
        if(lasty == 0)
        {
            LCD_SetPixel(LCD_MakePoint(i + 35, 229 - offset), LCD_YELLOW);
        }
        else
        {
            LCD_Line(LCD_MakePoint(i + 34, 229 - lasty), LCD_MakePoint(i + 35, 229 - offset), LCD_YELLOW);
        }
        lasty = offset;
    }
}

//Scan 1..30 MHz bandpass with 100 kHz step and display graph curve
void ScanFull(void)
{
    int i;
    //scan VSWR over the whole band with 100 kHz step
    FONT_Write(FONT_FRAN, LCD_PURPLE, LCD_BLACK, 180, 0, "Scanning pass 1            ");
    for(i = 0; i < nValFull; i++)
    {
        uint32_t freq = (1000 + i * 100) * 1000;
        DDS_SetFreq(freq);
        UADC adc = AD8302_ReadADC();
        swrvalues[i] = GetVSWR(adc);
        if (TOUCH_IsPressed())
        {
            FONT_Write(FONT_FRAN, LCD_PURPLE, LCD_BLACK, 180, 0, "                           ");
            return;
        }
    }

    FONT_Write(FONT_FRAN, LCD_PURPLE, LCD_BLACK, 180, 0, "                           ");
    //display VSWR graph
    LCD_DrawBitmap(LCD_MakePoint(0, 0), swr1_bmp, swr1_bmp_size);
    int lasty = 0;
    for(i = 0; i < nValFull; i++)
    {
        int offset = CalcOffset(swrvalues[i]);
        //zero point x = 15 y = 229
        if(i == 0)
        {
            LCD_SetPixel(LCD_MakePoint(i + 15, 229 - offset), LCD_GREEN);
        }
        else
        {
            LCD_Line(LCD_MakePoint(i + 14, 229 - lasty), LCD_MakePoint(i + 15, 229 - offset), LCD_GREEN);
        }
        lasty = offset;
    }
    ScanSmall();
}

extern "C"
int main()
{
    int i;
    SysTick_Config(SystemCoreClock / 1000);

    LCD_Init();
    TOUCH_Init();
    DDS_Init();
    AD8302_Init();

    DDS_SetFreq(10000 * 1000);

    LCD_FillAll(LCD_BLACK);
    FONT_Write(FONT_CONSBIG, LCD_RED, LCD_BLACK, 110, 60, "E U 3 E U");
    FONT_Write(FONT_FRANBIG, LCD_CYAN, LCD_BLACK, 30, 100, "antenna analyzer v 1.0");
    Sleep(1000);
    for(;;)
    {//main loop
        //Panoramic VSWR mode window.
        //To change it to measurement window,  tap the lower left part of display.
        //To pause, tap lower right part
        //To change center frequency of "small" scan, in step of +/- 1 MHz, tap top part of
        //the display, right or left from the center.
        for(;;)
        {
            ScanFull();
            if (TOUCH_IsPressed())
            {
                //Possibly there is a request to change frequency or to change window
                LCDPoint pt;
                if (TOUCH_Poll(&pt))
                {
                    if (pt.y > 120)
                    {//Tapped in lower part of the display
                        if (pt.x < 160)
                        {
                            //Left down angle tapped: change window
                            DDS_SetFreq(MeasurementFreq);
                            LCD_FillAll(LCD_BLACK);
                            FONT_Write(FONT_FRANBIG, LCD_WHITE, LCD_BLACK, 40, 60, "Measurement mode");
                            Sleep(250);
                            while(TOUCH_IsPressed());
                            break;
                        }
                        else
                        {
                            //Right down angle tapped: pause
                            FONT_Write(FONT_FRAN, LCD_RED, LCD_BLACK, 180, 0, "PAUSE            ");
                            while (TOUCH_IsPressed());
                            Sleep(200);
                            while (!TOUCH_IsPressed());
                            FONT_Write(FONT_FRAN, LCD_RED, LCD_BLACK, 180, 0, "                 ");
                            while (TOUCH_IsPressed());
                            continue;
                        }
                    }
                    if (pt.x > 160)
                    {
                        if (FMidMHz < 32)
                            FMidMHz++;
                    }
                    else
                    {
                        if (FMidMHz > 2)
                            FMidMHz--;
                    }
                    DrawFreqSmall();
                    Sleep(100);
                    continue;
                }
            }
            for (i = 0; i < 50; i++)
            {
                Sleep(50);
                if (TOUCH_IsPressed())
                    break;
            }
        }
        //Measurement mode window. To change it to VSWR tap the lower part of display.
        //To change frequency, in steps of +/- 500, 100 and 10 kHz, tap top part of the display,
        //the step depends on how far you tap from the center.
        LCD_FillAll(LCD_BLACK);
        for(;;)
        {
            ShowF();
            UADC adc = AD8302_ReadADC();
            ShowADC(adc);
            if (Touch())
            {
                LCD_FillAll(LCD_BLACK);
                FONT_Write(FONT_FRANBIG, LCD_WHITE, LCD_BLACK, 20, 60, "Panoramic VSWR mode");
                Sleep(250);
                while(TOUCH_IsPressed());
                break; //change mode to VSWR panoramic scan
            }
            for (i = 0; i < 4; i++)
            {
                Sleep(50);
                if (TOUCH_IsPressed())
                    break;
            }
        }
    }// for(;;) - main loop

    return 0;
}

#ifdef USE_FULL_ASSERT
void assert_failed(uint8_t* file, uint32_t line)
{
    for (;;);
}
#endif
